#include <AST.h>
#include <InferContext.h>
#include <MethodSet.h>
#include <decls.h>

#include "QualSolver.h"


void ConstructorDeclNode::infer( const InferContext &context )
{
  if (context.isExported)
    decl()->type()->paramTypes()->conservative( context.solver );
}


void DataDeclNode::infer( const InferContext &context )
{
  assert( decl()->type() == dtype() );
  
  dtype()->protect( context.solver );
  
  if (!initExpr()->absent())
    initExpr()->type()->weaklySubsumedBy( *dtype(), context.solver );
}


void FieldDeclNode::infer( const InferContext &context )
{
  if (context.isExported)
    dtype()->conservative( context.solver );

  DataDeclNode::infer( context );
}


void MethodNode::infer( const InferContext &context )
{
  const MethodDecl &myDecl = *decl();
  const TypeNode &myType = *myDecl.type();

  if (context.isExported || myDecl.container()->container()->fullName() == "ti.domains")
    myType.conservative( context.solver );

  {
    returnType()->protect( context.solver );
    
    const TypeNode * const thisType = myDecl.thisType();    
    if (thisType)
      thisType->protect( context.solver );
  }

  {
    const Decl * const overrides = myDecl.overrides();
    if (overrides)
      myType.identicalTo( *overrides->type(), context.solver );
  }

  {
    const MethodSet &implements = myDecl.implements();
    for (MethodSet::const_iterator interface = implements.begin();
	 interface != implements.end(); ++interface)
      myType.identicalTo( *(*interface)->type(), context.solver );
  }
}


void MethodDeclNode::infer( const InferContext &context )
{
  const MethodDecl &me = *decl();
  if (me.modifiers() & Native || isMain())
    me.type()->conservative( context.solver );
  
  MethodNode::infer( context );
}


void ParameterNode::infer( const InferContext &context )
{
  dtype()->protect( context.solver );
}
