#include <utility>

#include <AST.h>
#include <PhaseTimer.h>
#include <SolverInferContext.h>
#include <StaleBindings.h>
#include <compiler.h>
#include <st-field.h>
#include <template.h>
#include <typecheck.h>
#include <using-std.h>
#include <utils.h>

#include "../options.h"
#include "enact-context.h"
#include "options.h"
#include "solver.h"
#include "stats.h"
#include "traverse.h"


void inferLocal()
{
  if (opt_local) {
    PhaseTimer timer( "local qualification inference" );
    
    llist< TreeNode * > *allTrees = templateEnv.allInstances();
    foreach( unit, llist< CompileUnitNode * >, *allFiles )
      push( allTrees, (*unit) );
    
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->unshareTypeTree();

    LocalSolver solver;
    StaleBindings stale;
    LocalContext context( solver, stale );
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->inferLocalTree( context );

    // solver.dumpSolutions();

    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->localEnactTree( solver );

    free_all( allTrees );
    stale.rebindAll();
    
    // sanity checking
    TreeNode::TypeContext typeContext;
    foreach( tree, llist< CompileUnitNode * >, *allFiles )
      (*tree)->typecheck( &typeContext );
    
    if (debug_qual)
      LocalEnactContext::printStats( cerr );
  }
}
