#include <AST.h>
#include <PhaseTimer.h>
#include <SolverInferContext.h>
#include <StaleBindings.h>
#include <compiler.h>
#include <llist.h>
#include <template.h>
#include <typecheck.h>

#include "../options.h"
#include "enact-context.h"
#include "options.h"
#include "phase.h"
#include "solution.h"
#include "solver.h"


void inferSharing()
{
  if (opt_sharing) {
    PhaseTimer timer( "sharing qualification inference" );

    llist< TreeNode * > *allTrees = templateEnv.allInstances();
    foreach( unit, llist< CompileUnitNode * >, *allFiles )
      push( allTrees, (*unit) );
    
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->unshareTypeTree();

    SharingSolver solver;
    StaleBindings stale;
    
    SharingContext context( solver, stale );
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->inferSharingLateTree( context );

    if (sharingEnforcement == Early)
      foreach( tree, llist< TreeNode * >, *allTrees )
	(*tree)->inferSharingEarlyTree( solver );

    SharingSolution solution;
    solver.solve( solution );
    
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->sharingEnactTree( solution );

    free_all( allTrees );
    stale.rebindAll();

    // sanity checking
    TreeNode::TypeContext typeContext;
    foreach( tree, llist< CompileUnitNode * >, *allFiles )
      (*tree)->typecheck( &typeContext );
    
    if (debug_qual)
	SharingEnactContext::printStats( cerr );
  }
}
