#ifndef __MONITOR_H_
#define __MONITOR_H_

#ifdef HAVE_MONITORS

#ifdef COMM_AM2
	#include "monitor-dist.h"
#endif /* COMM_AM2 */

/* note: with the current definition of static, class monitors are useless
   (i.e. the synchronized keyword means nothing on a static method) */
# define MONITOR_FETCH_CLASS( staticstruct, holder )  TO_GLOBALB( (holder), 0, &STATIC_REF(staticstruct, monitor))
/* holder is always global, but object might be local */
#define MONITOR_FETCH_INSTANCE_LOCAL(  holder, object )  globalize( (holder), &((object)->monitor) )
#define MONITOR_FETCH_INSTANCE_GLOBAL( holder, object )  FIELD_ADDR_GLOBAL( holder, object, monitor )
#define MONITOR_LOCK(   holder )  monitor_enter( CAST_GPTR(holder) )
#define MONITOR_UNLOCK( holder )  monitor_exit(  CAST_GPTR(holder) )
#else /* no monitors */
# define MONITOR_FETCH_CLASS( staticstruct, holder )
# define MONITOR_FETCH_INSTANCE_LOCAL(  holder, object )
# define MONITOR_FETCH_INSTANCE_GLOBAL( holder, object )
# define MONITOR_LOCK(   holder )  
# define MONITOR_UNLOCK( holder ) 
#endif /* no monitors */


#ifdef CAUTIOUS_MONITORS
/* init monitor state that implements cautious waits */
#define CAUTIOUS_MONITOR_INIT(mon) {\
  mon->numwaiting = 0;\
  mon->numsignaled = 0;\
  }
#else
#define CAUTIOUS_MONITOR_INIT(mon)
#endif

#ifdef COMM_AM2
INLINE_MODIFIER void monitor_init(tic_monitor_t *mon)
INLINE({
  mon->proc = (Process)-1;
  mon->box = (Box)-1;

  mon->waitingProcs = NULL; /* no processes waiting to enter the monitor */
  mon->tryingProcs  = NULL; /* no processes trying to enter the monitor */
})

#else 

INLINE_MODIFIER void monitor_init(tic_monitor_t *mon)
INLINE({
  /* This lock controls access to the monitor structure */
/*
  __tic_lock_init(&(mon->lock));
  __tic_cond_init(&(mon->condvar), &(mon->lock));
*/
  mon->lock = __tic_lock_initializer;
  mon->condvar = __tic_cond_initializer;

  mon->proc = (Process)-1;
  mon->box = (Box)-1;

  mon->nesting = 0;
  mon->waitingProcs = NULL; /* no processes waiting to enter the monitor */
 
  CAUTIOUS_MONITOR_INIT(mon); /* init special fields if we're running CAUTIOUS_MONITORS */

})
#endif

INLINE_MODIFIER void simplemonitor_init(tic_simplemonitor_t *mon)
INLINE({
  mon->lock = __tic_lock_initializer;
  mon->condvar = __tic_cond_initializer;
  CAUTIOUS_MONITOR_INIT(mon); /* init special fields if we're running CAUTIOUS_MONITORS */
})

void monitor_startup();
void monitor_destroy(struct tic_monitor_t *monitor);
void monitor_enter(jGPointer monitor);
void monitor_exit(jGPointer monitor);
void monitor_wait(jGPointer monitor, ___tic_time_t *tm);
void monitor_notify(jGPointer monitor);
void monitor_notify_all(jGPointer monitor);

#endif /* __MONITOR_H_ */
