/* comm_utils.h - miscellaneous declarations & utilities */
/* see copyright.txt for usage terms */

#ifndef _COMM_UTILS_H
#define _COMM_UTILS_H

#define INIT_DONE = 4;

/* tic_monitor_t is used for language level monitors (objects, classes, etc). */
#ifdef COMM_AM2
	typedef struct proc_queue {
    		volatile Box box;
    		volatile Process proc;
    		volatile struct proc_queue *cdr;
	} proc_queue;

	typedef struct tic_monitor_t {
  		volatile Process proc;			/* Process that owns the monitor */
  		volatile Box box;			/* Box that the process lives on */
  		volatile proc_queue * waitingProcs; 	/* list of waiting processes */
  		volatile proc_queue * tryingProcs;	/* list of trying processes */
	} tic_monitor_t;
#else
	typedef struct waiting_proc {
    		Box box;
    		Process proc;
    		int nesting;
    		struct waiting_proc *next;
	} waiting_proc;

	typedef struct tic_monitor_t {
  		__tic_lock_t lock;
  		__tic_cond_t condvar;
  		volatile Process proc;
  		volatile Box box;
  		volatile int nesting;
  		struct waiting_proc * volatile waitingProcs;
  		#ifdef CAUTIOUS_MONITORS
    		  /* some pthread implementations have wait functions require some extra help */
    		  volatile int numwaiting;
    		  volatile int numsignaled;
  		#endif
	} tic_monitor_t;

#endif /* COMM_AM2 */

/* tic_simplemonitor_t is used for other special purposes in the runtime
 * and always uses real thread locks/condvars, regardless of backend
 */
        typedef struct tic_simplemonitor_t {
                __tic_lock_t lock;
                __tic_cond_t condvar;
                #ifdef CAUTIOUS_MONITORS
                  /* some pthread implementations have wait functions require some extra help */
                  volatile int numwaiting;
                  volatile int numsignaled;
                #endif
        } tic_simplemonitor_t;


typedef struct process_monitor_table {
  jGPointer monitor;
  unsigned int nesting;
  struct process_monitor_table *cdr;
} process_monitor_table;

struct boxInfo;

typedef struct _processInfo {
    Process processNumber;           		/* Process Number */
    Process boxProcessNumber;        		/* Process Number w.r.t. this box */
    struct boxInfo *box;             		/* The box this process is on */
    __tic_thread_t threadID;     		/* Local thread ID on this box */

    process_monitor_table *locked_monitors;    	/* List of monitors (PMT) locked by this thread */
    volatile int waiting_for_monitor;        	/* Current monitor this process is waiting for */
    volatile Process signal_proc;		/* Used to signal next proc for lock */
    volatile Box signal_box;			/* Corresponding box */

    volatile int barrier_phase;			/* Private state used to implement local barrier in barrier.c */

    char pad[256];                		/* Ensure each process struct lives on a separate cache line */
} processInfo;

typedef struct boxInfo {
  char *DNSname;                   /* DNS name of box */
  int pid;                         /* Process ID */
  int argc;                        /* Number of Arguments to Titanium */
  char **argv;                     /* Arguments to Titanium */
  Box boxNumber;                   /* Box Number */
  char *processesPerProcessors;    /* Processes per Processor */
  Process countProcesses;          /* Number of processes on this box */
  Process countProcessors;         /* Processors we want to use per box */
  Process totalProcessors;         /* Processors per box */
  processInfo *processes;          /* List of processes on this box */
} boxInfo;


/* Function Prototypes */

int COMM_Init_One();
int COMM_Init_Two(char *memory);
void COMM_Init_Three(char *memory);
void COMM_Init_Four(int argc, char **argv);
processInfo *COMM_GetHisBoxProcess(Process p) __attribute__ ((const));
processInfo *COMM_GetHisProcess(Process p) __attribute__ ((const));
processInfo *COMM_GetMyProcess() __attribute__ ((const));
juint COMM_GetMyGuid() __attribute__ ((const));
julong COMM_GetMyGuid64() __attribute__ ((const));
Box COMM_GetMyBoxNumber() __attribute__ ((const));
Box COMM_GetBoxNumberForProcNumber(Process processNumber) __attribute__ ((const));
Process COMM_GetHisBoxProcNumber(Process processNumber) __attribute__ ((const));
Process COMM_GetMyBParallelDegree() __attribute__ ((const));
Box COMM_GetBParallelDegree() __attribute__ ((const));
Process COMM_GetParallelDegree() __attribute__ ((const));
Process COMM_GetProxyProcNumberForBoxNumber(Box boxNumber) __attribute__ ((const));
float COMM_GetMyProcsPerProc() __attribute__ ((const));

#ifndef COMM_AM2
	void COMM_thread_cons_monitor(jGPointer monitor);
	void COMM_thread_remove_monitor(jGPointer monitor);
#endif

#endif /* _COMM_UTILS_H */
