/* main.c - Main program for tic backends */
/* see copyright.txt for usage terms */

#include <stdio.h>
#include <signal.h>
#include <assert.h>
#define TIC_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <tic.h>

/* get MAXHOSTNAMELEN */ 
#if defined(SOLARIS)
#include <netdb.h>
#else 
#include <sys/param.h>
#endif /* SOLARIS */

#if defined(COMM_AM2) && defined(HAVE_ENAMESERVER_H) 
#include <enameserver.h>
#endif

#include "main.h"
#include "ti-gc.h"
#include "native-utils.h"

int	___AM_MAX_SHORT;
int	___AM_MAX_MEDIUM;
int	___AM_MAX_LONG;

#if defined(COMM_AM2)
/* If request and reply bundles are the same, then share variable */
eb_t	__sc_bundle;
ep_t	__sc_endpoint;
en_t    __sc_endpoint_name;
char __sc_node_description[512]; /* for debugging purposes */
#endif /* AMII */
int usingAMSPMD = 0; /* true for workers using the AMUDP/AMMPI SPMD job startup API */

static char *progName;

int politep = 1;  /* True */

void free_resources(void);


/* return log2 if exact, else return -1 */
int getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}


#ifdef COMM_AM2
#ifdef HAVE_ENAMESERVER_H
#define NS_Safe(func) { if (func != 0) {fprintf(stderr, "ERROR: %s failed\n", #func); exit(1);}}

void setupNames()			   
{
  char 	inName[2048];
  byte	*result;
  int		result_len;
  int 	i;
  int		translations;
  en_t    *__sc_global = (en_t *) ti_malloc(sizeof(en_t)*COMM_GetBParallelDegree());

  AM_Safe(AM_MaxNumTranslations(&translations));
  if (translations < COMM_GetBParallelDegree()) {
    fprintf(stderr,
	    "ERROR: AM_MaxNumTranslations (%d) less than BOXES (%d).\n",
	    translations, COMM_GetBParallelDegree());
    exit(-1);
  }

  /* Insert our global endpoint name into the name server */
  if (PROCS == 1) {
    AM_Safe(AM_Map(__sc_endpoint, 0, __sc_endpoint_name, TIC_TAG));
  }
  else {
#ifdef GLUNIX
    NS_Safe(NameServer_Init());
#endif
    sprintf(__sc_node_description, 
	    "tic mill progname: %s global pid: %d mybox: %d", 
	    progName, COMM_GetMyGuid(), COMM_GetMyBoxNumber());
    NS_Safe (NameServer_Bind_Version_Flags
	     (__sc_node_description, (unsigned char*)&__sc_endpoint_name, sizeof (en_t),
	      ENS_VERSION_NOT_THERE,
	      ENS_FLAG_DESTROY_BINDING_ON_CONNECTION_CLOSE));
#ifdef GLUNIX
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#endif /* Glunix */
    for (i = 0; i < COMM_GetBParallelDegree(); i++) {
      if (COMM_GetMyBoxNumber() != i) {
	sprintf(inName, "tic mill progname: %s global pid: %d mybox: %d", 
		progName, COMM_GetMyGuid(), i);
	do {
	  NameServer_Lookup(inName, &result, &result_len);
	} while (result_len == 0);
	if (result == NULL) {
	  fprintf(stderr, "ERROR: result Null for lookup %s\n", inName);
	  exit(-1);
	}			
	if (result_len != sizeof(en_t)) {
	  fprintf(stderr, 
		  "ERROR: result_len(%d)!=sizeof(en_t)(%d) for lookup %s\n",
		  result_len, sizeof(en_t), inName);
	  exit(-1);
	}
	memcpy(&__sc_global[i], result, sizeof(en_t));
	AM_Safe(AM_Map(__sc_endpoint, i, __sc_global[i], TIC_TAG));
      }
    }
  }
}

#elif defined(COMM_AMLAPI)
int setupNames() {
  int i;
  int counter = 0;
	/* Array to store en_t address from all tasks */
  void **global_addr = (void**) ti_malloc(sizeof(void*)*COMM_GetBParallelDegree());
	/* Buffer of en_t address */
  en_t *buffer = (en_t *) ti_malloc(sizeof(en_t));
  int* buffer_test = (int *) ti_malloc (sizeof (int));

	/* Origin and Target Counter with address */
  void **tgt_addr = (void**) ti_malloc(sizeof(void*)*COMM_GetBParallelDegree());
  lapi_cntr_t  lapi_cntr;	/* Origin counter */

  lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */

	/* Initialize counters to be zero at the start */
  LAPI_Setcntr(*lapi_hndl, &lapi_cntr, 0);

	/* Exchange buffer address and target counter of every task */
  LAPI_Address_init (*lapi_hndl, &__sc_endpoint_name, global_addr);

  LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
  for (i = 0; i < COMM_GetBParallelDegree(); i++) {
    if (COMM_GetMyBoxNumber() != i) {
      LAPI_Get(*lapi_hndl, i, sizeof(en_t), (void*) global_addr[i],
			(void*) buffer, NULL, &lapi_cntr); 
      while( counter == 0 ) LAPI_Getcntr(*lapi_hndl, &lapi_cntr, &counter);
      LAPI_Setcntr(*lapi_hndl, &lapi_cntr, 0);
      counter = 0;
      /* printf ("%i thinks %i's endpoint name is: %i/%i and it's own endpoint name is %i/%i \n", AMLAPI_MyProc (), i, buffer->node, buffer->endpoint_id, __sc_endpoint_name.node, __sc_endpoint_name.endpoint_id); */

      AM_Safe(AM_Map(__sc_endpoint, i, *buffer, TIC_TAG));
    } 
    LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
  }
}
#else
void setupNames()
{
    fprintf(stderr,
            "ERROR: tried to access a global endpoint name server, but Tic was compiled with no nameserver\n"); 
    fflush(stderr);
    exit(1);
}
#endif
#endif /* COMM_AM2 */

#ifdef COMM_AM2

extern void bulkSetupHandlers();

char *___AM_SEGOFFSET = NULL;

void setupAM()
{
  uintptr_t maxSegLength;
  int 	maxHandlers;
  int		i;
	
  if (!usingAMSPMD) {
    AM_Safe(AM_Init());
    AM_Safe(AM_AllocateBundle(AM_SEQ, &__sc_bundle));
    AM_Safe(AM_AllocateEndpoint(__sc_bundle, &__sc_endpoint,
			        &__sc_endpoint_name));
  }
  AM_Safe(AM_MaxSegLength(&maxSegLength));
#ifdef USE_GC_NONE
  AM_Safe(AM_SetSeg(__sc_endpoint, (void *)NULL, maxSegLength));
  ___AM_SEGOFFSET = NULL;
#else
  AM_Safe(AM_SetSeg(__sc_endpoint, (void *)DATASTART, maxSegLength));
  ___AM_SEGOFFSET = DATASTART;
#endif
	
  /*	AM_Safe(AM_SetEventMask(__sc_bundle, AM_NOEVENTS)); */

  /* Setup Handlers */
  if ((maxHandlers = AM_MaxNumHandlers()) < __LAST_HANDLER) {
    fprintf(stderr, "ERROR: Only %d handler table entries available; need %d\n",
	    maxHandlers, __LAST_HANDLER);
    exit(1);
  }
	
  for (i = 0; ; i++) {
    if (handler_list[i].handler_index == __LAST_HANDLER) {
      break;
    }
    AM_Safe(AM_SetHandler(__sc_endpoint, handler_list[i].handler_index,
			  handler_list[i].func));
  }
		
  for (i = 0; ; i++) {
    if (reply_handler_list[i].handler_index == __LAST_HANDLER) {
      break;
    }
    AM_Safe(AM_SetHandler(__sc_endpoint,
			  reply_handler_list[i].handler_index,
			  reply_handler_list[i].func));
  }

  bulkSetupHandlers();
		
  if (!usingAMSPMD) 
    AM_Safe(AM_SetTag(__sc_endpoint, TIC_TAG));

  ___AM_MAX_SHORT = AM_MaxShort();
  ___AM_MAX_MEDIUM = AM_MaxMedium();
  ___AM_MAX_LONG = AM_MaxLong();
}

#endif /* COMM_AM2 */

int parseNumNodes(int *pargc, char ***pargv) {
  /* called by setup_global_layer to parse and remove the number of nodes 
   * argument for an implicit startup master 
   */
  int num_nodes;
  int i;

  if (*pargc < 2) {
    fprintf(stderr, "Tic: Missing parallel degree\n");
    fprintf(stderr, "Tic: Specify degree as first argument or use rexec/glurun/tcrun to start job\n");
    fprintf (stderr,"Tic: Usage '%s <num_nodes> {program arguments}'\n", (*pargv)[0]);
    exit(-1);
  }
  /*
   * argv[1] is number of nodes; argv[0] is program name; argv is
   * list of arguments including program name and number of nodes.
   * We need to remove argv[1] when the argument array is passed
   * to the tic_main().
   */
  num_nodes = atoi((*pargv)[1]);
  if (num_nodes < 1) {
    fprintf (stderr, "Tic: Invalid number of nodes: %s\n", (*pargv)[1]);
    fprintf (stderr, "Tic: Usage '%s <num_nodes> {program arguments}'\n", (*pargv)[0]);
    exit (1);
  }
  #ifdef GLUNIX
  { int i = Glib_GetNumAvailableNodes();
    if (i < num_nodes) {
      fprintf (stderr, "Tic: Error - Only %d nodes available (requested %d).\n",
               i, num_nodes);
      exit (1);
    }
  } 
  #endif	

  /* remove the num_nodes argument */
  for (i = 1; i < (*pargc)-1; i++) {
    (*pargv)[i] = (*pargv)[i+1];
  }
  (*pargv)[(*pargc)-1] = NULL;
  (*pargc)--;

  if (!getenv("TI_BACKEND_SILENT")) {
    printf("Tic: Spawning %s on %d nodes\n", (*pargv)[0], num_nodes);
    fflush(stdout);
  }

  return num_nodes;
}

extern julong gpid;

#ifdef COMM_AMUDP
#ifdef REXEC 
    TI_IDENT(AMUDP_DEFAULT_SPAWNFN_IDENT_STRING, "$TitaniumDefaultSpawnFunction: R $");
#elif defined(GLUNIX)
    TI_IDENT(AMUDP_DEFAULT_SPAWNFN_IDENT_STRING, "$TitaniumDefaultSpawnFunction: G $");
#else /* AMUDP implicit ssh startup */       
    TI_IDENT(AMUDP_DEFAULT_SPAWNFN_IDENT_STRING, "$TitaniumDefaultSpawnFunction: S $");
#endif                                                                                                    
#endif

void setup_global_layer(int *pargc, char ***pargv) {

/* this function is responsible for performing "implicit" job startup.
 * That is, spawning a parallel job when the application was executed
 * directly without the use of a job spawner like glurun, rexec, tcrun, etc.
 * It also takes advantage of AMUDP's flexible job startup API where appropriate
 * The command-line format for all implicit startups is:
 * <programname> <paralleldegree> {program args...}
 */

#if defined(GLUNIX) || defined(COMM_AMUDP) || defined(COMM_AMMPI)
  /* we only support implicit startup for these job spawning systems */
#else
  return;
#endif

#ifdef COMM_AMMPI
 {/* AMMPI startup is always explicit, but we'll handle it here */
  int networkdepth=0;

  #if 0
  {
  int initialized = 0;
  char systemErrDesc[MPI_MAX_ERROR_STRING+10];
  int len = MPI_MAX_ERROR_STRING;
  int errval;

  /* Init MPI, if necessary (this may result in job spawn, depending on the MPI implementation) */
  if ((errval=MPI_Initialized(&initialized)) != MPI_SUCCESS) {
    if (MPI_Error_string(errval, systemErrDesc, &len) != MPI_SUCCESS || len == 0) 
      strcpy(systemErrDesc, "(no description available)");
    fprintf(stderr,"Error calling MPI_Initialized(): %s\n", systemErrDesc); 
    fflush(stderr);
    abort();
  }
  if (!initialized) {
    if ((errval=MPI_Init(pargc, pargv)) != MPI_SUCCESS) {
      if (MPI_Error_string(errval, systemErrDesc, &len) != MPI_SUCCESS || len == 0) 
        strcpy(systemErrDesc, "(no description available)");
      fprintf(stderr,"Error calling MPI_Initialized(): %s\n", systemErrDesc); 
      fflush(stderr);
      abort();
    }
  }
  }
  #endif

  AMMPI_VerboseErrors = 1; /* get detailed info about any failures */
  if (getenv("TI_BACKEND_SILENT")) {
    AMMPI_SilentMode = 1; /* suppress diagnostic output */
  }
  if (getenv("TI_NETWORKDEPTH")) 
    networkdepth = atoi(getenv("TI_NETWORKDEPTH"));
  if (networkdepth < 1) networkdepth = 20; /* a reasonable default for many systems */

  /* startup this worker */
  { int rank;
    AM_Safe(AMMPI_SPMDStartup(pargc, pargv, networkdepth, &gpid, &__sc_bundle, &__sc_endpoint));
    rank = AMMPI_SPMDMyProc();
    AM_Safe(AM_GetTranslationName(__sc_endpoint, rank, &__sc_endpoint_name)); /* lookup our name */
    usingAMSPMD = 1; /* signal that we've completed endpoint initialization */
    return;
  }
 } 
#endif

/* generic initialization */
#ifdef GLUNIX
  if (!(Glib_Initialize() != 0)) {
    fprintf(stderr, "Global layer initialization failed.\n");
    fflush(stderr);
    exit(-1);
  }	
#endif /* Glunix */

#ifdef COMM_AMUDP
  AMUDP_VerboseErrors = 1; /* get detailed info about any failures */
  if (getenv("TI_BACKEND_SILENT")) {
    AMUDP_SilentMode = 1; /* suppress diagnostic output */
  }
  if (AMUDP_SPMDIsWorker(*pargv)) { /* we're an implicit worker using AMUDP SPMD API */
    int rank;
    AM_Safe(AMUDP_SPMDStartup(pargc, pargv, 0, 0, NULL, &gpid, &__sc_bundle, &__sc_endpoint));
    rank = AMUDP_SPMDMyProc();
    AM_Safe(AM_GetTranslationName(__sc_endpoint, rank, &__sc_endpoint_name)); /* lookup our name */
    usingAMSPMD = 1; /* signal that we've completed endpoint initialization */
    if (getenvMaster("TI_BACKEND_SILENT")) {
      AMUDP_SilentMode = 1; /* suppress diagnostic output */
    }
  }
  else { /* decide whether we're an implicit master */
    int isExplicitWorker = 0;
    char spawnfn = '\0';
#ifdef REXEC 
    isExplicitWorker = (int)getenv("REXEC_PAR_DEGREE");
    spawnfn = 'R';
#elif defined(GLUNIX)
    isExplicitWorker = !Glib_AmIStartup();
    spawnfn = 'G';
#else /* AMUDP implicit ssh startup */
    isExplicitWorker = 0; /* we don't support explicit ssh startup */
    spawnfn = 'S';
#endif
    if (!isExplicitWorker) { /* we are an implicit master */
      spawnfn_t fp = NULL;
      int networkdepth = 0;
      int num_nodes = parseNumNodes(pargc, pargv);

      if (getenv("TI_SPAWNFN")) /* user has overridden default implicit job spawner */
        spawnfn = *getenv("TI_SPAWNFN");
      if (getenv("TI_NETWORKDEPTH")) 
        networkdepth = atoi(getenv("TI_NETWORKDEPTH"));
      if (networkdepth < 1) networkdepth = 20; /* a reasonable default for many systems */

      switch (spawnfn) {
#ifdef GLUNIX
        case 'g': case 'G':
          fp = AMUDP_SPMDGlunixSpawn;
          break;
#endif
        case 's': case 'S':
          fp = AMUDP_SPMDSshSpawn;
          break;
#ifdef REXEC
        case 'r': case 'R':
          fp = AMUDP_SPMDRexecSpawn;
          break;
#endif
        case 'l': case 'L':
          fp = AMUDP_SPMDLocalSpawn;
          break;
        case 'c': case 'C':
          fp = AMUDP_SPMDCustomSpawn;
          break;
        default:
          fprintf (stderr, "Tic: Invalid spawn function specified in TI_SPAWNFN\n");
          fprintf (stderr, "Tic: The following mechanisms are available:\n");
          fprintf (stderr, 
            "    'S'  Spawn jobs using ssh remote shells\n"
            "    'L'  Spawn jobs using fork()/exec() on the local machine (good for SMP's)\n"
          #ifdef REXEC
            "    'R'  Spawn jobs using rexec (Berkeley Millennium only)\n"
          #endif
          #ifdef GLUNIX
            "    'G'  Spawn jobs using GLUNIX (Berkeley NOW only)\n"
          #endif
            "    'C'  Spawn jobs using custom job spawner (TI_CSPAWN_CMD)\n"
            );
          exit (1);
        }
    AM_Safe(AMUDP_SPMDStartup(pargc, pargv, num_nodes, networkdepth, fp, &gpid, NULL, NULL));
    abort(); /* master startup should never return */
    }
  }
#else /* not COMM_AMUDP */
#ifdef GLUNIX
  if (Glib_AmIStartup()) {
    int num_nodes = parseNumNodes(pargc, pargv);
      /* Spawn only returns if it failed */
    if (Glib_Spawn(num_nodes, (*pargv)[0], *pargv)) {
      perror("Global spawn failed");
      exit(-1);
    }
    abort();
  }
#endif /* Glunix */
#endif
}

#define TIENV_FILENAME ".tienv"
/* read and parse the .tienv file in the executable directory for optional 
 * environment initialization (required for some distributed spawners that don't properly
 * propagate environment variables to the slave processes). Input is program pathname.
 */
void readEnvVarFile(char *progpathname) {
  char filename[255];
  char line[1024];
  int linenum = 0;
  char *p;
  FILE *fp;
  strcpy(filename, progpathname); /* look for file in executable directory */
  if (strchr(filename,'/')) p = strrchr(filename,'/');
  else p = strrchr(filename,'\\');
  if (p) *(p+1) = '\0';
  else filename[0] = '\0';
  strcat(filename, TIENV_FILENAME);
  fp = fopen(filename, "r");
  if (!fp) return;

  while (fgets(line, 1024, fp)) {
    char *p = line;
    char *pend = NULL;
    char delim;
    char name[255];
    char value[1024];
    int sawequal=0;
    linenum++;
    while (*p && isspace(*p)) p++;
    if (!*p || *p == '#') continue; /* # starts a comment */
    if (*p == '"') { p++; pend = p + strcspn(p, "\""); } /* get name */
    else pend = p + strcspn(p, "\t =");
    strncpy(name, p, pend-p); name[pend-p] = '\0';
    if (*pend && *pend != '=') p = pend+1;
    else p = pend;
    while (*p && isspace(*p)) p++;
    if (*p == '=') { sawequal=1; p++; }
    while (*p && isspace(*p)) p++;
    if (*p == '"') { p++; pend = p + strcspn(p, "\""); } /* get name */
    else pend = p + strcspn(p, "\t ");
    strncpy(value, p, pend-p); value[pend-p] = '\0';
    if (!*name || !sawequal) {
      fprintf(stderr,"Tic: Ignoring illegal environment initializer at %s:%i.\n"
                     "Tic: format is one variable per line: name = value\n"
                     "Tic: where name and value may be \"quoted\" and name must be non-empty.\n",
                     filename, linenum);
    }
    else { 
      if (setenv(name, value, 1)) 
        perror("setenv: in readEnvVarFile()");
    }
  } 
  if (ferror(fp)) perror("fgets()");

  if (!getenv("TI_BACKEND_SILENT"))
    printf("Tic: Reading environment variable settings from %s\n", filename);
}

void
free_resources ()
{
#ifdef COMM_AM2 
  if (usingAMSPMD) {
#if defined(COMM_AMUDP) 
    AMUDP_SPMDExit(0);
#elif defined(COMM_AMMPI)
    AMMPI_SPMDExit(0);
#endif
    }
  else if (COMM_GetBParallelDegree() > 1) {
    AM_Safe(AM_FreeEndpoint (__sc_endpoint));
    AM_Safe(AM_FreeBundle (__sc_bundle));
    AM_Safe(AM_Terminate());
#ifdef HAVE_ENAMESERVER_H
    NS_Safe((NameServer_Finish(),0)); /* NameServer_Finish() has a void return */
#endif
  }
#endif /* AMII */
}

static int init_complete = 0;
void
free_resource_handler (int sig)
{
    if (init_complete) printExitStats(0);
    free_resources ();
#ifdef GLUNIX
  /* Glunix needs this to prevent orphaned processes */
  Glib_Signal(Glib_GetMyNpid(), VNN_ALL, SIGTERM);
#endif
    exit (0);
}

__tic_thread_key_t tic_thread_key;



int ___PROCS;
int ___BOXES;
Box ___MYBOX;

#ifdef MEMORY_SHARED
int ___MYBOXPROCS;
#else
Process ___MYPROC;
#endif

void memory_init();
void stats_init(void);

void init_box_consts() {
  /* This process is a node in a parallel job. */
  ___PROCS = COMM_GetParallelDegree();
  ___BOXES = COMM_GetBParallelDegree();
  ___MYBOX = COMM_GetMyBoxNumber();
  #ifdef MEMORY_SHARED
    ___MYBOXPROCS = COMM_GetMyBParallelDegree();
  #else
    ___MYPROC = myProcessNumber();
  #endif

  com_init();
  store_init();
  memory_init();
  #if defined(USE_DISTRIBUTED_GC) && !defined(USE_GC_NONE)
    dgc_init();
  #endif
  stats_init();

  __tic_lock_init(&AM_lock);
}

extern char *RUNTIME_MODEL;
char *RUNTIME_MODEL = TIC_BACKEND_NAME; 

/* these strings are read from the executable by ident and tcrun */
TI_IDENT(BackendIdent, "$TitaniumBackend: " TIC_BACKEND_NAME " $");

extern void region_init(void);

void *tic_main(int argc, char **argv)
{
  ti_main(argc, argv);
  return 0;
}


void *thread_go(void *procinfo) {
  processInfo *procInfo = (processInfo *)procinfo;
  int result = __tic_thread_set_key_value(tic_thread_key, (void *)procInfo);
  if (result != 0) {
    printf("error in tic thread set key value: %d\n", result);
    return NULL;
  }

  #ifdef _CRAY
    /* Setup a trap handler for SIGFPE (floating point exception) to keep it from crashing us */
    sigignore(SIGFPE);
  #endif

  #ifdef DEBUG_THREADS
    fprintf(stderr,"Processor starting MYPROC=%i PROCS=%i MYBOX=%i BOXES=%i MYBOXPROC=%i MYBOXPROCS=%i\n", 
      MYPROC, PROCS, MYBOX, BOXES, MYBOXPROC, MYBOXPROCS); 
    fflush(stderr);
  #endif

  local_barrier(); /* wait for all local threads */

  region_init();
 
  monitor_startup();
 
  barrier();
  init_complete = 1;

  ti_main (procInfo->box->argc, procInfo->box->argv);

  barrier();

  printExitStats(1);

  return NULL;
}

void spawn_threads() {
  int i;
  processInfo *thisproc;
  
  __tic_thread_key_init(&tic_thread_key);

  barrier_init(); /* this must happen on the primordial thread before thread spawn */

#ifdef MEMORY_DISTRIBUTED
  /* need an external barrier here to allow titanium barrier to be bootstrapped on all nodes */
#if defined(COMM_AMUDP)
  if (usingAMSPMD) AMUDP_SPMDBarrier(); 
  else 
#elif defined(COMM_AMMPI)
  if (usingAMSPMD) AMMPI_SPMDBarrier(); 
  else 
#endif
  {
#if defined(GLUNIX)
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#elif defined( COMM_AMLAPI )
  lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */
  LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
#else
  abort(); /* we need something else here - barrier() is no good */
  barrier(); /* this is not strictly correct and may cause problems */
#endif /* Glunix */
    }
#endif

  /* this thread will eventually act as thread 0 for this box */
  thisproc = COMM_GetHisBoxProcess(0);
  __tic_thread_set_key_value(tic_thread_key, thisproc);
  thisproc->threadID = __tic_thread_self();

  init_box_consts(); /* called on proc 0 before other threads spawned */

  /* spawn other threads */
  for (i = 1; i < COMM_GetMyBParallelDegree(); i++) {
    int result;
    thisproc = COMM_GetHisBoxProcess(i);
    __tic_create_thread(thread_go, (void *)thisproc, &thisproc->threadID, result);
    if (result != 0) {
      perror("tic_create_thread");
      exit(-1);
    }
    #ifdef DEBUG_THREADS
      printf("created thread 0x%08x\n",(int) thisproc->threadID);
    #endif
  }
  
  thisproc = COMM_GetHisBoxProcess(0);
  thread_go(thisproc); /* run process 0 */
  return;
}

void join_threads() {
  int i;
  
    for (i = 1; i < COMM_GetMyBParallelDegree(); i++) {
      int result;
      processInfo *thisproc = COMM_GetHisBoxProcess(i);

      #ifdef DEBUG_THREADS
        printf("joining thread 0x%08x\n",(int) thisproc->threadID);
      #endif

      result = __tic_thread_join(thisproc->threadID, NULL);

      if (result != 0) {
	perror("tic_join_threads");
	exit(-1);
      }
    }
}

static volatile int tic_frozen = 1;
static int _frozen() { /* ensure a full stack frame */
  return *&tic_frozen;
}
void freezeForDebugger(char *hostname) {
  printf("frozen for debugger: host %s, pid %d\n", hostname, getpid());
  fflush(stdout);
  while (_frozen()) tic_sched_yield();
}

int real_main (int argc, char** argv, char** envp);
int main (int argc, char** argv, char** envp) {
  int dummy = 0;
  ti_gc_stack_init(&dummy); /* called to set the cold stack ptr for the primordial thread */
  return real_main(argc, argv, envp);
}
int real_main (int argc, char** argv, char** envp) {
    int countBoxMem, countProcMem;
    char *boxMem, *procMem;
    char	hostname[MAXHOSTNAMELEN] = "(unknown)";

    progName = argv[0];

    readEnvVarFile(progName);

    /* do this *very* early, in case library routines allocate memory */
    ti_gc_mem_init(); /* init the GC for the main thread - not necessary for dynamically created threads, should only be run on the primordial thread */

    signal (SIGHUP, free_resource_handler);
    signal (SIGINT, free_resource_handler);
    signal (SIGQUIT, free_resource_handler);
    signal (SIGTERM, free_resource_handler);

    setup_global_layer(&argc, &argv); /* must come before any worker-related code, as this may be the startup master node */

    countBoxMem = COMM_Init_One();
    boxMem = (char *) tic_alloca(countBoxMem); /* try to get thread-affinity */

    gethostname(hostname, sizeof(hostname));
    
    if (getenvMaster("TI_FREEZE")) freezeForDebugger(hostname);

    countProcMem = COMM_Init_Two(boxMem);
    procMem = (char *) tic_alloca(countProcMem); /* try to get thread-affinity */

    COMM_Init_Three(procMem);

#if defined(MEMORY_DISTRIBUTED) || defined(MEMORY_SHARED)
  if (!getenvMaster("TI_BACKEND_SILENT")) {
    printf("Tic: %s is box number %d of %d\n", hostname, 
	 COMM_GetMyBoxNumber(), COMM_GetBParallelDegree());
    fflush(stdout);
    }
#endif

    if (COMM_GetBParallelDegree() > 1) {
#ifdef COMM_AM2
      setupAM();
#else /* COMM_AM2 */
      fprintf(stderr, "Error: Trying to use more than one box without a known communications layer.\n");
      return 2;
#endif /* COMM_AM2 */
    }
    
    COMM_Init_Four(argc, argv);
    
    if (COMM_GetBParallelDegree() > 1) {
#ifdef COMM_AM2
      if (usingAMSPMD) {
#if defined(COMM_AMUDP)
        AM_Safe(AMUDP_SPMDBarrier()); /* need a barrier after handler setup */
#elif defined(COMM_AMMPI)
        AM_Safe(AMMPI_SPMDBarrier()); /* need a barrier after handler setup */
#endif
        }
      else setupNames();
#if defined(COMM_AMUDP) || defined(COMM_AMMPI)
    /* AMUDP/AMMPI requires us to specify the number of peers and the desired network depth */
    if (!usingAMSPMD) { /* SPMD startup already takes care of this */
      int networkdepth = 0;
      if (getenv("TI_NETWORKDEPTH")) 
        networkdepth = atoi(getenv("TI_NETWORKDEPTH"));
      if (networkdepth < 1) networkdepth = 20; /* a reasonable default for many systems */
      AM_Safe(AM_SetExpectedResources(__sc_endpoint, COMM_GetBParallelDegree(), networkdepth));
    }
#endif
#else /* COMM_AM2 */
      fprintf(stderr, "Error: Trying to use more than one box without a known communications layer.\n");
      return 2;
#endif /* COMM_AM2 */
    }

    __tic_thread_set_concurrency(COMM_GetMyBParallelDegree());

    spawn_threads();
    
    join_threads();

    fflush(stdout);
    fflush(stderr);

#ifdef GLUNIX
    /* don't free AM endpoints until we're sure everyone is ready (AM-based barrier() not good enough) */
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#endif /* Glunix */

    free_resources ();
    
    return 0; /* if Titanium program returned from void main, that means return 0 */
}

