#ifndef _include_bulk_h_
#define _include_bulk_h_

#include <string.h>


#define local_to_global_copy( local, global, count )				\
  do {										\
    if (isDirectlyAddressable( global ))					\
      memcpy(   TO_LOCAL( global ), (local), sizeof(*(local)) * (count) );	\
    else									\
      bulk_write( (global), (local), sizeof(*(local)) * (count) );	\
  } while (0)


#define global_to_local_copy( global, local, count )				\
  do {										\
    if (isDirectlyAddressable( global ))					\
      memcpy(   (local), TO_LOCAL( global ), sizeof(*(local)) * (count) );	\
    else									\
      bulk_read( (local),  (global), sizeof(*(local)) * (count), tic_has_ptrs );	\
  } while (0)

#ifdef COMM_AM2
  /* AM-based scatter-gather primitives (for internal use in ti_array.c only) */

  /* rectangular scatter/gather */
  void get_array(void *pack_method, void *copy_desc, int copy_desc_size, 
	         int tgt_box, void *buffer, int atomicelements);
  void put_array(void *unpack_method, void *copy_desc, int copy_desc_size,
	         void *array_data, int array_data_size, int tgt_box);

  /* sparse scatter/gather - transfer data back & forth between a 
   * local, packed data list and the memory space of a remote proc
   * using a list of addresses on the remote proc
   * requires: sizeof(data_list) == num_elem * elem_sz
   *           sizeof(remote_addr_list) == num_elem * sizeof(void*)
   */
  void sparse_scatter(void **remote_addr_list, void *src_data_list, 
                      int remote_box, int num_elem, int elem_sz, int atomic_elements);
  void sparse_gather(void *tgt_data_list, void **remote_addr_list, 
                      int remote_box, int num_elem, int elem_sz, int atomic_elements);
#endif

#endif /* !_include_bulk_h_ */
