/* handlers.c - Install AM handlers */
/* see copyright.txt for usage terms */

#include <tic.h>

#ifdef COMM_AM2
#if defined(AMNOW)
extern void __error_handler(int status, op_t opcode, void *argblock);
#endif
#if defined(COMM_AMUDP) || defined(COMM_AMMPI)
extern void ___amstats_request_handler(void *token, void *buf, int nbytes, int procnum, int reserved);
#endif


handler_func_t reply_handler_list[] = {
{	__B_GET_RH,		___b_get_reply_handler		},
{	__SH_GET_RH,	        ___sh_get_reply_handler		},
{	__I_GET_RH,		___i_get_reply_handler		},
{	__F_GET_RH,		___f_get_reply_handler		},
{	__D_GET_RH,		___d_get_reply_handler		},
{	__L_GET_RH,	        ___l_get_reply_handler		},
{	__LP_GET_RH,	        ___lp_get_reply_handler		},
{       __GP_GET_RH,            ___gp_get_reply_handler		},
{	__DECR_CTR_RH,		__decr_ctr_reply_handler	},
{	__DECR_PUT_CTR_RH,	__decr_put_ctr_reply_handler	},
{	__GETCOMPLETE_RH,	___get_complete_reply_handler	},
{	__PUTCTRCOMPSRC_RH,	___put_ctr_comp_src_reply_handler},
{	__NULL_RH,		__null_reply_handler		},
{	__MON_ENTER_RH,         ___monitor_enter_reply_handler	},
{	__MON_EXIT_RH,          ___monitor_exit_reply_handler	},
{	__MON_WAIT_RH,          ___monitor_wait_reply_handler	},
{	__MON_SIGNAL_RH,	___monitor_signal_reply_handler	},
{	__MON_NOTIFY_RH,        ___monitor_notify_reply_handler	},
{	__LAST_HANDLER,		NULL				}
};

handler_func_t handler_list[] = {
#if defined(AMNOW)
{	__ERROR_H,		(void (*)())__error_handler	},
#endif
{	__B_GET_H,		___b_get_handler		},
{	__SH_GET_H,		___sh_get_handler		},
{	__I_GET_H,		___i_get_handler		},
{	__F_GET_H,		___f_get_handler		},
{	__D_GET_H,		___d_get_handler		},
{	__L_GET_H,		___l_get_handler		},
{	__LP_GET_H,		___lp_get_handler		},
{	__GP_GET_H,             ___gp_get_handler		},
{	__B_PUT_H,		___b_put_handler		},
{	__SH_PUT_H,		___sh_put_handler		},
{	__I_PUT_H,		___i_put_handler		},
{	__F_PUT_H,		___f_put_handler		},
{	__D_PUT_H,		___d_put_handler		},
{	__L_PUT_H,	        ___l_put_handler		},
{	__LP_PUT_H,	        ___lp_put_handler		},
{	__GP_PUT_H,             ___gp_put_handler		},
{	__BULK_PUT_H,		___bulk_put_handler		},
{	__BULK_STORE_H,		___bulk_store_handler		},
{	__B_STORE_H,		___b_store_handler		}, 
{	__SH_STORE_H,		___sh_store_handler		},
{	__I_STORE_H,		___i_store_handler		},
{	__F_STORE_H,		___f_store_handler		},
{	__D_STORE_H,		___d_store_handler		},
{	__L_STORE_H,		___l_store_handler		},
{	__LP_STORE_H,		___lp_store_handler		},
{	__GP_STORE_H,		___gp_store_handler		},
{	__BARRIER_INCR_H,	__barrier_incr_handler		},
{	__GET_H,		___get_handler			},
{	__MON_ENTER_H,          ___monitor_enter_request_handler	},
{	__MON_EXIT_H,           ___monitor_exit_request_handler		},
{	__MON_WAIT_H,           ___monitor_wait_request_handler		},
{	__MON_SIGNAL_H,		___monitor_signal_request_handler	},
{	__MON_NOTIFY_H,         ___monitor_notify_request_handler	},
{	__MON_NOTIFY_ALL_H,     ___monitor_notify_all_request_handler	},
{       __MON_CANCEL_WAIT_H,	___monitor_cancel_wait_request_handler	},
#if defined(COMM_AMUDP) || defined(COMM_AMMPI)
{	__AMSTATS_H,		___amstats_request_handler	},
#endif
{	__LAST_HANDLER,		NULL				}
};

#if defined(COMM_AMNOW)
#include <am2/arg_block.h>

void __error_handler(int status, op_t opcode, arg_block_t *argblock)
{
        int i;

        printf("AM layer failed on node %d.\n", myProcess());

        printf ("opcode = %d\n", argblock->opcode);
        printf ("request endpoint: %#X\n", argblock->request_endpoint);
        printf ("reply endpoint: %d\n", argblock->reply_endpoint);
        printf ("dest offset: %d\n", argblock->dest_offset);
        printf ("source addr: %#X\n", argblock->source_addr);
        printf ("nbytes: %d\n", argblock->nbytes);
        printf ("handler: %d\n", argblock->handler);
        for (i = 0; i < 8; i++)
            printf ("data[%d]: %d\n", i, argblock->data[i]);

        switch (status) {
          case EBADARGS:
                printf("Invalid arguements to request or reply\n");
                break;
          case EBADENTRY:
                printf("Unbound table entry\n");
                break;
          case EBADTAG:
                printf("Sender Tag did not match Receiver Tag\n");
                break;
          case EBADHANDLER:
                printf("Invalid index into receiver's handler table\n");
                break;
          case EBADSEGOFF:
                printf("Invalid offset into destination memory segment\n");
                break;
          case EBADLENGTH:
                printf("Bulk transfer length exceeds sengment length\n");
                break;
          case EBADENDPOINT:
                printf("Destination endpoint does not exist\n");
                break;
          case ECONGESTION:
                printf("Persistent congestion at destination\n");
                break;
          case EUNREACHABLE:
                printf("Unreachable destination\n");
                break;
          default:
                printf("Unknown status value: %d\n", status);
        }

        free_resources ();

        fflush(stdout);
        exit(-1);
}
#endif /* COMM_AMNOW */

#endif /*AMII*/
	
	
	

