/* store.c - signaling global memory accesses (store) */
/* see copyright.txt for usage terms */

#include <tic.h>

__tic_lock_t	Store_mutex;

void store_init(void)
{
  if (MYBOXPROC == 0) {
    __tic_lock_init(&Store_mutex);
  }
}

/* .....................
 *
 * Keep track of STORESs
 *
 * .....................
 */

Counter sentBytes = 0;
Counter recvBytes = 0;

#ifdef COMM_AM2
/* Wait until all stores from all processors have completed */
void all_store_sync(void)
{
	int res;

	barrier();
	do {
		res = all_reduce_to_all_iadd(recvBytes-sentBytes);
	} while(res != 0);

	__tic_lock(&Store_mutex);
	recvBytes = sentBytes = 0;
	__tic_unlock(&Store_mutex);
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	barrier();
}
#else
#define all_store_sync()
#endif /* AMII */

#ifdef COMM_AM2
void all_store_sync_ctr(Counter *rbytes, Counter *sbytes)
{
	int res;

	barrier();
	do {
		res = all_reduce_to_all_iadd(*rbytes - *sbytes);
	} while(res != 0);

	__tic_lock(&Store_mutex);
	*rbytes = *sbytes = 0;
	__tic_unlock(&Store_mutex);
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	barrier();
}
#else 
#define all_store_sync_ctr(a, b)
#endif /* AMII */

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */


#ifdef COMM_AM2

___STORE_HANDLER4(b, jbyte)
___STORE_HANDLER4(sh, jshort)
___STORE_HANDLER4(i, jint)
___STORE_HANDLER4(f, jfloat)
___STORE_HANDLER8(d, jdouble)
___STORE_HANDLER8(l, jlong)
#ifdef PTR64
     ___STORE_HANDLER16(gp, jGPointer, jUIntPointer)
     ___STORE_HANDLER8(lp, void *)
#else
void ___gp_store_handler(void *token, juint addr, juint box, juint proc, juint addrData,
			    juint ctr) {
  jGPointer data;
  mutateglobalwithbox(data, proc, box, addrData);
  __gp_store_handler(token, (jGPointer *)addr, data, (Counter *)ctr);
}
     ___STORE_HANDLER4(lp, void *)
#endif

#endif /* COMM_AM2 */

#ifdef COMM_AM2
void __b_store_handler (void *token, jbyte *addr, jbyte data, Counter *ctr)
{
  addr[0] = data;
  __incr_ctr_size(ctr, sizeof(jbyte));
  TI_AM_Reply0(token, __NULL_RH);
}

void __sh_store_handler(void *token, jshort *addr, jshort data, Counter *ctr)
{
	addr[0] = data;
	__incr_ctr_size(ctr, sizeof(jshort));

	TI_AM_Reply0(token, __NULL_RH);
}

void __i_store_handler(void *token, jint *addr, jint data, Counter *ctr)
{
	addr[0] = data;
	__incr_ctr_size(ctr, sizeof(jint));

	TI_AM_Reply0(token, __NULL_RH);
}

void __f_store_handler(void *token, jfloat *addr, jfloat data, Counter *ctr)
{
  addr[0] = data;
  __incr_ctr_size(ctr, sizeof(jfloat));
  
  TI_AM_Reply0(token, __NULL_RH);
}


void __d_store_handler(void *token, jdouble *addr, jdouble data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jdouble));
  TI_AM_Reply0(token, __NULL_RH);
}

void __l_store_handler(void *token, jlong *addr, jlong data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jlong));
  TI_AM_Reply0(token, __NULL_RH);
}


void __lp_store_handler(void *token, void **addr, void *data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(void *));
  TI_AM_Reply0(token, __NULL_RH);
}


#ifdef PTR64
void __gp_store_handler(void *token, jGPointer *addr, jUIntPointer data0,
		      jUIntPointer data1, Counter *ctr)
{
  *addr = mutateglobalwithbox(*addr, boxproctoproc(data0), 
			      boxproctobox(data0), (void *)data1);
  __incr_ctr_size(ctr, sizeof(jGPointer));
  TI_AM_Reply0(token, __NULL_RH);
}
#else
void __gp_store_handler(void *token, jGPointer *addr,
			jGPointer data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jGPointer));
  TI_AM_Reply0(token, __NULL_RH);
}
#endif



#endif /* AMII */

