
// Uses fake templates.  Script replaces V_ARITY with specified.

package ti.domains;

//
// ========================
// Multi Rect Domain Type A
// ========================
//
// This represents a domain that is a list of Rect Domains.  It
// is optimized for a fast "foreach" at the expense of set comparisons
// and operations.
//

public class tiMultiRectADomainV_ARITY extends tiDomainV_ARITY {

  // Constructors
  public tiMultiRectADomainV_ARITY() {
    rects = new (tiDomains.region) tiRectDomainListV_ARITY();
  }
  public tiMultiRectADomainV_ARITY(tiRectDomainListV_ARITY rectList) {
    rects = rectList;
  }
  public tiMultiRectADomainV_ARITY(tiRectDomainV_ARITY rd) {
    rects = new (tiDomains.region) tiRectDomainListV_ARITY();
    if (!rd.isNull()) {
      rects.append(rd);
    }
  }
  public tiMultiRectADomainV_ARITY(tiRectDomainV_ARITY [1d] rda) {
    // unchecked: assert that RectDomain's in rda are disjoint
    rects = new (tiDomains.region) tiRectDomainListV_ARITY();
    foreach (p in rda.domain()) {
      rects.append(rda[p]);
    }
  }
  public tiMultiRectADomainV_ARITY(Point<V_ARITY> [1d] pa) {
    // unchecked: assert that Point's in pa are disjoint
    rects = new (tiDomains.region) tiRectDomainListV_ARITY();
    foreach (p in pa.domain()) {
      tiRectDomainV_ARITY rd =
	new tiRectDomainV_ARITY(pa[p],
				pa[p] + Point<V_ARITY>.all(1),
				Point<V_ARITY>.all(1));
      rects.append(rd);
    }
  }

  // Internal set relations routines
  private tiRectDomainListV_ARITY newSubtract(tiRectDomainListV_ARITY dl) {
    tiRectDomainListV_ARITY result = rects;

    for (tiRectDomainListIteratorV_ARITY d_iter = dl.newIterator();
	 !d_iter.isEnd();
	 d_iter.advance()) {
      tiRectDomainListV_ARITY saved_result = result;
      result = new (tiDomains.region) tiRectDomainListV_ARITY();
      for (tiRectDomainListIteratorV_ARITY iter = saved_result.newIterator();
	   !iter.isEnd();
	   iter.advance()) {

	tiRectDomainV_ARITY id =
	  iter.dereference().multiply(d_iter.dereference());

	if (id.isNull()) {
	  result.append(iter.dereference());
	} else if (id.isEqual(iter.dereference())) {
	  // do nothing
	} else {
	  tiRectDomainV_ARITY rd = iter.dereference();
	  Point<V_ARITY> p0 = rd.getLowerBound();
	  Point<V_ARITY> p1 = rd.getUpperBound();
	  Point<V_ARITY> stride = rd.getStride();

	  // slice
	  for (int i = 0; i < rd.getArity(); i++) {
	    int rd_p0 = p0.get(i);
	    int rd_p1 = p1.get(i);
	    int id_p0 = id.getLowerBound().get(i);
	    int id_p1 = id.getUpperBound().get(i);

	    if (rd_p0 < id_p0) {
	      result.append(new tiRectDomainV_ARITY(p0,
						    p1.set(i, id_p0),
						    stride));
	      p0 = p0.set(i, id_p0);
	    }
	    if (rd_p1 > id_p1) {
	      result.append(new tiRectDomainV_ARITY(p0.set(i,
							   (id_p1 -
							    1 +
							    stride.get(i))),
						    p1,
						    stride));
	      p1 = p1.set(i, id_p1);
	    }
	  }

	  // handle interior of overlap
	  for (int i = 0; i < rd.getArity(); i++) {
	    if (stride.get(i) < id.getStride().get(i)) {
	      for (int set_p0 = p0.get(i);
		   set_p0 < (p1.get(i) -
			     id.getStride().get(i));
		   set_p0 += id.getStride().get(i)) {
		result.
		  append(new tiRectDomainV_ARITY(p0.set(i,
							(set_p0 +
							 stride.get(i))),
						 p1.set(i,
							(set_p0 +
							 id.
							 getStride().
							 get(i))),
						 stride));
	      }
	      stride = stride.set(i,
				  id.getStride().get(i));
	    }
	  }
	}
      }
    }

    return result;
  }

  // Set Relations
  public tiDomainV_ARITY add(tiDomainV_ARITY d) {
    tiRectDomainListV_ARITY mdl = d.getRectangles();
    tiRectDomainListV_ARITY result = newSubtract(mdl);

    for (tiRectDomainListIteratorV_ARITY d_iter = mdl.newIterator();
	 !d_iter.isEnd();
	 d_iter.advance()) {
      result.append(d_iter.dereference());
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }
  public tiDomainV_ARITY difference(tiDomainV_ARITY d) {
    tiRectDomainListV_ARITY mdl = d.getRectangles();

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(newSubtract(mdl));
  }
  public tiDomainV_ARITY multiply(tiDomainV_ARITY d) {
    // this * d

    tiRectDomainListV_ARITY result = new (tiDomains.region) tiRectDomainListV_ARITY();
    tiRectDomainListV_ARITY mdl = d.getRectangles();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      for (tiRectDomainListIteratorV_ARITY d_iter = mdl.newIterator();
	   !d_iter.isEnd();
	   d_iter.advance()) {

	tiRectDomainV_ARITY partial_result =
	  iter.dereference().multiply(d_iter.dereference());

	if (!partial_result.isNull()) {
	  result.append(partial_result);
	}
      }
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }
  public boolean isNull() {
    // this is null

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      if (!iter.dereference().isNull()) {
	return false;
      }
    }
    return true;
  }
  public inline boolean isNotNull() {
    return !isNull();
  }
  public boolean isADR() {
    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
       !iter.isEnd();
       iter.advance()) {
      if (!iter.dereference().isADR()) {
      return false;
      }
    }
    return true;
  }
  public boolean isEqual(tiDomainV_ARITY d) {
    // this == d
    return (isSubset(d) && d.isSubset(this));
  }
  public boolean isNotEqual(tiDomainV_ARITY d) {
    return (!isSubset(d) || !d.isSubset(this));
  }
  public boolean isSubset(tiDomainV_ARITY d) {
    // this <= d
    tiRectDomainListV_ARITY mdl = d.getRectangles();
    return newSubtract(mdl).isEmpty();
  }
  public boolean isStrictSubset(tiDomainV_ARITY d) {
    // this < d
    return isSubset(d) && (!d.isSubset(this));
  }
  public boolean isSuperset(tiDomainV_ARITY d) {
    return d.isSubset(this);
  }
  public boolean isStrictSuperset(tiDomainV_ARITY d) {
    return d.isStrictSubset(this);
  }

  // RectDomain Relations
  public tiDomainV_ARITY add(tiRectDomainV_ARITY rd) {
    return add(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public tiDomainV_ARITY difference(tiRectDomainV_ARITY rd) {
    return difference(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public tiDomainV_ARITY multiply(tiRectDomainV_ARITY rd) {
    return multiply(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }

  public boolean isEqual(tiRectDomainV_ARITY rd) {
    return isEqual(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public boolean isNotEqual(tiRectDomainV_ARITY rd) {
    return isNotEqual(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public boolean isSubset(tiRectDomainV_ARITY rd) {
    return isSubset(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public boolean isStrictSubset(tiRectDomainV_ARITY rd) {
    return isStrictSubset(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public boolean isSuperset(tiRectDomainV_ARITY rd) {
    return isSuperset(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }
  public boolean isStrictSuperset(tiRectDomainV_ARITY rd) {
    return isStrictSuperset(new (tiDomains.region) tiMultiRectADomainV_ARITY(rd));
  }

  // Point Relations
  public tiDomainV_ARITY add(Point<V_ARITY> p) {
    // this + p
    tiRectDomainListV_ARITY result = new (tiDomains.region) tiRectDomainListV_ARITY();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      result.append(iter.dereference().add(p));
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }
  public tiDomainV_ARITY difference(Point<V_ARITY> p) {
    // this - p
    tiRectDomainListV_ARITY result = new (tiDomains.region) tiRectDomainListV_ARITY();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      result.append(iter.dereference().difference(p));
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }
  public tiDomainV_ARITY divide(Point<V_ARITY> p) {
    // this / p
    tiDomainV_ARITY result = new (tiDomains.region) tiMultiRectADomainV_ARITY();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      // Call the add function to remove any overlaps generated by rounding
      // effects during the divide.
      result = result.add(iter.dereference().divide(p));
    }

    return result;
  }
  public tiDomainV_ARITY multiply(Point<V_ARITY> p) {
    // this * p
    tiRectDomainListV_ARITY result = new (tiDomains.region) tiRectDomainListV_ARITY();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      result.append(iter.dereference().multiply(p));
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }

  public tiDomainV_ARITY permute(Point<V_ARITY> p) {
    tiRectDomainListV_ARITY result = new (tiDomains.region) tiRectDomainListV_ARITY();

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      result.append(iter.dereference().permute(p));
    }

    return new (tiDomains.region) tiMultiRectADomainV_ARITY(result);
  }

  // Shape Information
  public boolean doesContain(Point<V_ARITY> p) {
    // this contains p

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      if (iter.dereference().doesContain(p)) {
	return true;
      }
    }
    return false;
  }
  public boolean isRectangular() {
    joinRectangles();

    return rects.hasSingularity();
  }
  public int getNumPoints() {
    int result = 0;

    for (tiRectDomainListIteratorV_ARITY iter = rects.newIterator();
	 !iter.isEnd();
	 iter.advance()) {
      result = result + iter.dereference().getNumPoints();
    }

    return result;
  }
  public Point<V_ARITY> getLowerBound() {
    Point<V_ARITY> result;
    tiRectDomainListIteratorV_ARITY iter = rects.newIterator();

    if (iter.isEnd()) {
      return Point<V_ARITY>.all(0);
    } else {
      result = iter.dereference().getLowerBound();
      for (iter.advance();
	   !iter.isEnd();
	   iter.advance()) {
	result = result.getLowerBound(iter.dereference().getLowerBound());
      }
      return result;
    }
  }
  public Point<V_ARITY> getUpperBound() {
    Point<V_ARITY> result;
    tiRectDomainListIteratorV_ARITY iter = rects.newIterator();

    if (iter.isEnd()) {
      return Point<V_ARITY>.all(0);
    } else {
      result = iter.dereference().getUpperBound();
      for (iter.advance();
	   !iter.isEnd();
	   iter.advance()) {
	result = result.getUpperBound(iter.dereference().getUpperBound());
      }
      return result;
    }
  }
  public Point<V_ARITY> getUpperMaxPoint() {
    return getUpperBound().sub(Point<V_ARITY>.all(1));
  }
  public tiRectDomainV_ARITY getBoundingBox() {
    return new tiRectDomainV_ARITY(getLowerBound(),
				   getUpperBound(),
				   Point<V_ARITY>.all(1));
  }
  public tiRectDomainListV_ARITY getRectangles() {
    return rects;
  }
  public tiRectDomainV_ARITY [1d] getRectDomains() {
    return (tiRectDomainV_ARITY [1d])getRectDomainsLocal();
  }
  public tiRectDomainV_ARITY [1d] local getRectDomainsLocal() {
    int numRectangles = rects.getNumItems();
    tiRectDomainV_ARITY [1d] local result =
      new tiRectDomainV_ARITY [[1 : numRectangles : 1]];
    int numRectsAdded = 0;

    for (tiRectDomainListIteratorV_ARITY iter_x = rects.newIterator();
         !iter_x.isEnd();
         iter_x.advance()) {
      numRectsAdded++;
      result[[numRectsAdded]] = iter_x.dereference();
    }
    // assert (numRectsAdded == numRectangles);
    return result;
  }
  public Point<V_ARITY> [1d] getPoints() {
    return (Point<V_ARITY> [1d])getPointsLocal();
  }
  public Point<V_ARITY> [1d] local getPointsLocal() {
    int numPoints = getNumPoints();
    Point<V_ARITY> [1d] local result =
      new Point<V_ARITY> [[1 : numPoints : 1]];
    int numPointsAdded = 0;

    for (tiRectDomainListIteratorV_ARITY iter_x = rects.newIterator();
         !iter_x.isEnd();
         iter_x.advance()) {
      RectDomain<V_ARITY> rd = [iter_x.dereference().getLowerBound() :
                                iter_x.dereference().getUpperMaxPoint() :
                                iter_x.dereference().getLoopingStride()];
      foreach (p in rd) {
	numPointsAdded++;
	result[[numPointsAdded]] = p;
      }
    }
    // assert (numPointsAdded == numPoints);
    return result;
  }

  // Internal dynamic optimization routines
  private void joinRectangles() {
    boolean changes_made;
    tiRectDomainListV_ARITY currList = rects;

    do {
      changes_made = false;
      tiRectDomainListV_ARITY newList = new (tiDomains.region) tiRectDomainListV_ARITY();
      boolean ignore[] = new (tiDomains.region) boolean[currList.getNumItems()];

      for (int i = 0; i < currList.getNumItems(); i++) {
	ignore[i] = false;
      }

      int searchIndexInit = 0;
      for (tiRectDomainListIteratorV_ARITY iter_x = currList.newIterator();
	   !iter_x.isEnd();
	   iter_x.advance()) {
	if (!ignore[searchIndexInit]) {
	  int searchIndex = searchIndexInit + 1;
	  boolean matchFound = false;
	  for (tiRectDomainListIteratorV_ARITY iter_comp_x =
		 currList.newIterator(iter_x).advance();
	       ((!iter_comp_x.isEnd()) && (!matchFound));
	       iter_comp_x.advance()) {
	    if (!ignore[searchIndex]) {
	      tiDomainV_ARITY result =
		iter_x.dereference().add(iter_comp_x.dereference());
	      tiRectDomainListV_ARITY rdl = result.getRectangles();
	      if (rdl.hasSingularity()) {
		tiRectDomainV_ARITY rd = rdl.getSingularity();

		newList.append(rd);
		changes_made = true;
		matchFound = true;
		ignore[searchIndex] = true;
	      }
	    }
	    searchIndex++;
	  }
	  if (!matchFound) {
	    newList.append(iter_x.dereference());
	  }
	}
	searchIndexInit++;
      }
      currList = newList;
    } while (changes_made);
    rects = currList;
  }

  // Dynamic optimizations
  public tiDomainV_ARITY demote() {
    joinRectangles();

    return this;
  }

  // Convert to a string
  public String toString() {
    String result;
    boolean isFirst;

    isFirst = true;
    result = "";
    for (tiRectDomainListIteratorV_ARITY d_iter = rects.newIterator();
	 !d_iter.isEnd();
	 d_iter.advance()) {
      if (!isFirst) {
	result = result + " + ";
      }
      result = result + d_iter.dereference().toString();
      isFirst = false;
    }

    return result;
  }

  // Debugging
  /*
  public void print(String desc) {
    int count = 0;
    for (tiRectDomainListIteratorV_ARITY d_iter = rects.newIterator();
	 !d_iter.isEnd();
	 d_iter.advance()) {
      count++;
      tiRectDomainV_ARITY rd = d_iter.dereference();
      Point<V_ARITY> p = rd.getLowerBound();
      System.out.print(desc);
      System.out.print("(");
      System.out.print(count);
      System.out.print(") [[");
      for (int i = 0; i < p.arity(); i++) {
	if (i != 0) System.out.print(",");
	System.out.print(p.get(i));
      }
      System.out.print("] : ");
      Point<V_ARITY> p2 = rd.getUpperBound();
      System.out.print("[");
      for (int i = 0; i < p2.arity(); i++) {
	if (i != 0) System.out.print(",");
	System.out.print(p2.get(i));
      }
      System.out.print("] : ");
      Point<V_ARITY> p3 = rd.getStride();
      System.out.print("[");
      for (int i = 0; i < p3.arity(); i++) {
	if (i != 0) System.out.print(",");
	System.out.print(p3.get(i));
      }
      System.out.println("]]");
    }
  }
  */

  // State
  private tiRectDomainListV_ARITY rects;

}
