#define PAGE_GROUP_SIZE 32
#define K 4

__tic_lock_t freepages_lock;
#ifdef REGIONPROFILE
static __tic_lock_t rplock;
#endif

#ifdef REGIONPROFILE
#define RPLOCK __tic_lock(&rplock)
#define RPUNLOCK __tic_unlock(&rplock)

jIntPointer totalbytes, maxbytes;
jIntPointer totalregions, maxregions;
#endif

jIntPointer allocation_pagesize;

void (*ti_gc_other_roots_pusher)();

#ifdef ALIGN
#undef ALIGN
#endif
#define ALIGN(x, n) ((x) + ((n) - 1) & ~((n) - 1))
#define RALIGNMENT 8

#include "ti-gc.h"
#include "regions.h"
#include <pages.c>
#include <alloc.c>


FILE_STATIC_STATIC_DEF(rstart, int);

FILE_STATIC_STATIC_DEF(allregions, Region);

/* Doubly linked region list management */
void addregion(Region r)
/* Effects: Adds p to the front of doubly-linked region list */
{
  r->previous = NULL;
  r->next = FILE_STATIC_STATIC_REF(allregions);
  if (FILE_STATIC_STATIC_REF(allregions)) FILE_STATIC_STATIC_REF(allregions)->previous = r;
  FILE_STATIC_STATIC_REF(allregions) = r;
}

void remregion(Region r)
/* Effects: Removes r from the doubly-linked region list */
{
  if (r->previous)
    r->previous->next = r->next;
  else
    FILE_STATIC_STATIC_REF(allregions) = r->next;
  if (r->next)
    r->next->previous = r->previous;
}

void clear(void *start, jIntPointer size)
{
  long *clear, *clearend;

  clear = (long *)start;
  clearend = (long *)((char *)start + size);
  do *clear++ = 0;
  while (clear < clearend) ;
}

Region newregion(void)
{
  char *first = (char *)alloc_single_page(NULL);
  Region r;

#ifdef REGIONPROFILE
  RPLOCK;
  totalregions++;
  if (totalregions > maxregions) maxregions = totalregions;
  RPUNLOCK;
#endif

  /* stagger regions across cache lines a bit */
  FILE_STATIC_STATIC_REF(rstart) += 64;
  if (FILE_STATIC_STATIC_REF(rstart) > RPAGESIZE / K) FILE_STATIC_STATIC_REF(rstart) = 0;
  r = (Region)(first + FILE_STATIC_STATIC_REF(rstart) + offsetof(struct page, previous));
  clear(r, sizeof *r);

  return r;
}

void initregion(Region r)
{
  char *first =
    (char *)r - FILE_STATIC_STATIC_REF(rstart) - offsetof(struct page, previous);
  int fnb = PAGENB(first);

  /* Start using page with region header as a pointer-containing page */
  r->normal.page.base = first;
  /* XXX: Need extra ALIGNs ? */
  r->normal.page.allocfrom = ALIGN(FILE_STATIC_STATIC_REF(rstart) + offsetof(struct page, previous) + sizeof(*r), RALIGNMENT);

  /* Guarantee failure for all other blocks */
  r->normal.superpage.allocfrom = K * RPAGESIZE + 1;
  r->normal.hyperpage.allocfrom = K * K * RPAGESIZE + 1;
  r->atomic.page.allocfrom = RPAGESIZE + 1;
  r->atomic.superpage.allocfrom = K * RPAGESIZE + 1;
  r->atomic.hyperpage.allocfrom = K * K * RPAGESIZE + 1;

  /* Remember that r owns this page. */
  r->normal.pages = (struct page *)first;
  set_region(r->normal.pages, 1, r);

  /* Add to list of all regions */
  addregion(r);
}

void *ralloc(Region r, size_t size)
{
  char *mem;

  if (!r) return ti_malloc(size);

#ifdef REGIONPROFILE
  RPLOCK;
  totalbytes += size;
  if (totalbytes > maxbytes) maxbytes = totalbytes;
  RPUNLOCK;
  r->bytes += size;
#endif

  size = ALIGN(size, RALIGNMENT);
  mem = (char *)qalloc(r, &r->normal, size);
  clear(mem, size);

  return mem;
}

const void *ralloc_atomic(Region r, size_t size)
{
  if (!r) return ti_malloc_atomic(size);
  else
    {
#ifdef REGIONPROFILE
      RPLOCK;
      totalbytes += size;
      if (totalbytes > maxbytes) maxbytes = totalbytes;
      RPUNLOCK;
      r->bytes += size;
#endif
      return qalloc(r, &r->atomic, ALIGN(size, RALIGNMENT));
    }
}

#if 0
static void delpage(char *deleting, char *end)
{
  while (deleting < end)
    {
      cleanup_t cln = GET_CLEANUP_FN(deleting);
      size_t s;

      if (!cln) break;
      s = cln(deleting);
      deleting += ALIGN(s, RALIGNMENT);
    }
}
#endif

static void delregion(Region r)
{
  char *p;
  unsigned long next, pnb;

#ifdef REGIONPROFILE
  RPLOCK;
  totalbytes -= r->bytes;
  totalregions--;
  RPUNLOCK;
#endif
#if 0
  /* Mark end of last page */
  if (h->normal.allocfrom < RPAGESIZE)
    *(void **)((char *)h->normal.lastpage + h->normal.allocfrom) = NULL;

  /* Scan normal pages */
  /* First page (last in list) is special (region object there) */
  for (pnb = PAGENB(h->normal.lastpage); next = nextmap[pnb]; pnb = next)
    {
      p = PAGEADDR(pnb);
      delpage(p, p + RPAGESIZE);
    }

  /* We skip region on first page */
  p = PAGEADDR(pnb);
  delpage((char *)(h + 1), p + RPAGESIZE);
#endif

  /* Remove r from region list */
  remregion(r);

  __tic_lock(&freepages_lock);

  free_all_pages(r, &r->normal);
  free_all_pages(r, &r->atomic);

  __tic_unlock(&freepages_lock);

}

int deleteregion(Region hp)
{
  delregion(hp);
  return 1;
}

#ifdef BACKEND_UNIPROC
/* dummy region for NULL and GC addresses. It has id 0 */
struct T6Region4lang2ti zeroregion;
#endif

void m4initmT6Region4lang2ti(void)
{
  /* Dummy fn so that this file gets included. It used to contain
     what became region_init... */
}

#ifdef REGIONPROFILE
static void print_memory_usage(void)
{
  Region r;

  if (MYPROC != 0)
    return;

  RPLOCK;
  fprintf(stderr, "Current region usage %lu bytes, max %lu bytes\n",
	  totalbytes, maxbytes);
  fprintf(stderr, "Current regions %lu, max %lu\n",
	  totalregions, maxregions);
  {
#if defined(MEMORY_SHARED)
  /* Assumes all threads are stopped */
  int i;

  for (i = 0; i < PROCS; i++)
    for (r = allregions[i]; r; r = r->next)
#else
  for (r = allregions; r; r = r->next)
#endif
    fprintf(stderr, "region %x %d used\n", r, r->bytes);
  }
  RPUNLOCK;
}
#endif

void *bounded_sub(void *p, IntPtr n)
{
  if ((IntPtr)p < n) return 0;
  else return (void *)((IntPtr)p - n);
}

void *bounded_add(void *p, IntPtr n)
{
  if ((IntPtr)p > (IntPtr)-n) return 0;
  else return (void *)((IntPtr)p + n);
}

void region_init(void)
{
  int dummy_local;
  static int dummy_global;
  void *start, *end;
  int i;

#ifndef USE_GC_NONE
  GC_setpagesize();
  allocation_pagesize = GC_page_size;
#endif

  if (RPAGESIZE > allocation_pagesize) allocation_pagesize = RPAGESIZE;
  /* Our minimum allocation chunk is K pages. We want to be sure that
     this is a multiple of the minimum allocation size (if it isn't,
     either K or RPAGESIZE should be increased).
     Ideally, RPAGESIZE <= GC_page_size <= K * RPAGESIZE
     (otherwise we may waste some memory because all allocation requests
     are rounded to allocation_pagesize) */
  assert(K * RPAGESIZE >= allocation_pagesize);

  /* CM: No longer needed. The page table region map is set to UNK_REGIONID
     as the default now. */
#if 0
  /* Initialise stack & static data page -> RegionId tables to contain
     UNK_REGIONID. This is somewhat hackish:
     - we assume that the stack will range from -64 megs to +1 megs from
       it's current position
     - we assume all static data is found within one meg of dummy_global
     Correct solutions will be very machine specific
  */
  set_region_range(bounded_sub(&dummy_local, 64 * 1024 * 1024),
		   bounded_add(&dummy_local, 1 * 1024 * 1024),
		   UNK_REGIONID);
  set_region_range(bounded_sub(&dummy_global, 1 * 1024 * 1024),
		   bounded_add(&dummy_global, 1 * 1024 * 1024),
		   UNK_REGIONID);
  /* null */
  set_region_range(0, (void *)1, UNK_REGIONID);
#endif

  init_pages();

  if (__tic_lock_init(&freepages_lock) < 0)
    {
      perror("ti_mutex_init");	
      exit(1);
    }
#ifdef REGIONPROFILE
  if (__tic_lock_init(&rplock) < 0)
    {
      perror("ti_mutex_init(rp)");	
      exit(1);
    }
#endif

#ifdef BACKEND_UNIPROC
  zeroregion.id = 1; /* Shared ! */
#endif

#ifdef REGIONPROFILE
  atexit(print_memory_usage);
#endif

  /* Initialize the region map or table */
#ifndef LARGE_ADDRESSES
  for (i = 0; i < MAXPAGE; i++) {
    __regionmap[i] = UNK_REGIONID;
  }
#else /* !LARGE_ADDRESSES */
  for (i = 0; i < (1 << MEMSLICE2); i++) {
    __regiontable[i] = (RegionId *) NULL;
  }
#endif /* !LARGE_ADDRESSES */

}

/* GC support. Mark all non-atomic pages */

void GC_push_region_roots(void)
{
  Region r;

#if defined(MEMORY_SHARED)
  /* Assumes all threads are stopped */
  int i;

  for (i = 0; i < PROCS; i++)
    for (r = allregions[i]; r; r = r->next)
#else
  for (r = allregions; r; r = r->next)
#endif
    {
      struct page *p;

      for (p = r->normal.pages; p; p = p->next)
	ti_region_push_all(p, (char *)p + RPAGESIZE);
      for (p = r->normal.bigpages; p; p = p->next)
	ti_region_push_all(p, (char *)p + p->pagecount * RPAGESIZE);
    }

  if (ti_gc_other_roots_pusher)
    ti_gc_other_roots_pusher();
}

void sharingViolation(void)
{
  fprintf(stderr, "Attempt to assign a reference to a private object into a shared or GC object\n");
  fflush(stderr);
  abort();
}
