/* ti_array.h */
/* This file assumes T is a type and N is an arity. */

#include "runtime-options.h"
#include "native-utils.h"
#include "assertExists.h"

/* Instructions on how to add a method are in ti_array.c */

#define N_LARGER_THAN_1 N_MINUS_1

#ifndef GLOBAL_ARRAY
#define GLOBAL_ARRAY 1
#endif

#ifndef DEBUG_ARRAYS
#define DEBUG_ARRAYS 0
#endif

#ifndef DUMPABLE_ARRAYS
#define DUMPABLE_ARRAYS 0
#endif

#ifndef DEFAULT_DUMP_DEST
#define DEFAULT_DUMP_DEST "|/disks/orodruin/titanium/work/pike/bin/aview"
#endif

#ifndef EXPLICITLY_STORE_CREATOR
#define EXPLICITLY_STORE_CREATOR 0
#endif

#if GLOBAL_ARRAY
  #define LOCAL_PART(p)	(TO_LOCAL(p))
  #define BOX_PART(p)	(TO_BOX(p))
  #define PROC_PART(p)	(TO_PROC(p))
  #define INDEX		INDEX_GLOBAL
#else
  #define LOCAL_PART(p) (p)
  #define BOX_PART(p)	MYBOX
  #define PROC_PART(p)  MYPROC
  #define INDEX		INDEX_LOCAL
#endif

#if GLOBAL_ARRAY && !GLOBAL_ELEMENTS
# define assignable 0
#else
# define assignable 1
#endif

#define PRINT_POINT(p, f) 					\
do {								\
  jint _i;							\
  fprintf(f, "[");						\
  for (_i = 1; _i <= N; _i++)					\
    fprintf(f, (_i > 1) ? ", %d" : "%d", POINT_INDEX(p, _i));	\
  fprintf(f, "]");						\
} while (0)

#define PRINT_DOMAIN(R, f)			\
do {						\
  jint _i;					\
  ti_POINT p = RECTDOMAIN_STRIDE(R);		\
  fputs("[", f);				\
  PRINT_POINT(RECTDOMAIN_MIN(R), f);		\
  fputs(" : ", f);				\
  PRINT_POINT(RECTDOMAIN_UPB(R), f);		\
  for (_i = 1; _i <= N; _i++)			\
    if (POINT_INDEX(p, _i) >= 2) break;		\
  if (_i <= N) {				\
    fputs(" : ", f);				\
    PRINT_POINT(p, f);				\
  }						\
  fputs("]", f);				\
} while (0)


/* Address of arr[p] is:
   desc.A + ((p[1] - desc.base[0]) / desc.stride[0]) * desc.sideFactors[0]
          + ((p[2] - desc.base[1]) / desc.stride[1]) * desc.sideFactors[1]
          + ...
*/

/* Functions defined in ti_array.cc */
extern void ti_dump(ti_ARRAY *px, char *dest, char *userArg);
extern ti_ARRAY ti_empty();
extern jboolean ti_isnull();
extern PTR_TO_T ti_get_array_data_ptr(const ti_ARRAY x);
extern ti_ARRAY ti_construct(const char *where, Region pregion, ti_RECTDOMAIN R, int isShared);
#if N_LARGER_THAN_1
extern ti_ARRAY_SLICE SLICE_METHOD(const ti_ARRAY x, jint k, jint j);
#else
void SLICE_METHOD(const ti_ARRAY x, jint k, jint j);
#endif
#if assignable
  extern void EXCHANGE_METHOD(const ti_ARRAY x, T w);
  extern void COPY_METHOD(const ti_ARRAY x, const ti_ARRAY y);
  void SET_METHOD(const ti_ARRAY x, T w);
  void COPY_WITHDOMAIN_METHOD(const ti_ARRAY x, const ti_ARRAY y, const ti_DOMAIN dom);
  void COPY_WITHRECTDOMAIN_METHOD(const ti_ARRAY x, const ti_ARRAY y, const ti_RECTDOMAIN rdom);
  void COPY_WITHPTARRAY_METHOD(const ti_ARRAY x, const ti_ARRAY y, const ti_1DARRAY_OF_POINT_N ptArray);
  ti_ARRAY MAKELOCALANDCONTIGUOUS_METHOD(const ti_ARRAY x);
  void SCATTER_METHOD(const ti_ARRAY destArray, const ti_1DARRAY_OF_T packedArray, const ti_1DARRAY_OF_POINT_N ptArray);
  void GATHER_METHOD(const ti_ARRAY srcArray, const ti_1DARRAY_OF_T packedArray, const ti_1DARRAY_OF_POINT_N ptArray);
#endif
ti_ARRAY TRANSLATE_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY PERMUTE_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY INJECT_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY PROJECT_METHOD(ti_ARRAY x, ti_POINT p);
jboolean ISCONTIGUOUS_METHOD(const ti_ARRAY x);
jboolean ISCONTIGUOUSOVERDOMAIN_METHOD(const ti_ARRAY x, const ti_RECTDOMAIN R);
void ti_printdomain(ti_ARRAY x, FILE *f);
void ti_boundscheck(const char *where, const ti_ARRAY *x, ti_POINT p);

/* bulk I/O functions */
#ifndef __bulkio_typedefs
#define __bulkio_typedefs
  /* these three seem to always get generated */
  #include "layout!T16RandomAccessFile2io4java.h"
  #include "layout!T15DataInputStream2io4java.h"
  #include "layout!T16DataOutputStream2io4java.h"

  /* these three are more of a struggle */
  #ifdef _include_layout_PT16RandomAccessFile2io4java_h_
    #define GP_RAF PT16RandomAccessFile2io4java
  #else
    typedef T16RandomAccessFile2io4java _RAF;
    typedef GP_type( _RAF ) GP_RAF;
  #endif
  #ifdef _include_layout_PT15DataInputStream2io4java_h_
    #define GP_DIS PT15DataInputStream2io4java
  #else
    typedef T15DataInputStream2io4java _DIS;
    typedef GP_type( _DIS ) GP_DIS;
  #endif
  #ifdef _include_layout_PT16DataOutputStream2io4java_h_
    #define GP_DOS PT16DataOutputStream2io4java
  #else
    typedef T16DataOutputStream2io4java _DOS;
    typedef GP_type( _DOS ) GP_DOS;
  #endif
#endif

#if ELEMENTS_ARE_ATOMIC
  #if assignable
    void READFROMRAF_METHOD(const ti_ARRAY x, GP_RAF fileobj);
    void READFROMDIS_METHOD(const ti_ARRAY x, GP_DIS fileobj);
  #endif
  void WRITETORAF_METHOD(const ti_ARRAY x, GP_RAF fileobj);
  void WRITETODOS_METHOD(const ti_ARRAY x, GP_DOS fileobj);
#endif

#if BOUNDS_CHECKING
jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p);
#else /* !BOUNDS_CHECKING */
INLINE_MODIFIER jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p)
{
 register jint sum = 0;
#if N_LARGER_THAN_1
  jint i;
#else
#define i 1
#endif

  assertExists(*x, where);
#if N_LARGER_THAN_1
  for (i = N; i >= 1; i--)
#endif
    {
      /* At this point, x->stride[i-1] should divide qm.  Of course, we only
	 know for sure if bounds checking is on. */
      const jint str = x->stride[i-1];
      const jint q = POINT_INDEX(p, i);
      const jint qm = q - x->base[i-1];
      register jint delta = qm * x->sideFactors[i-1];

      if (PREDICT_FALSE(str != 1)) delta /= str; /* only do the expensive divide when necessary */
      sum += delta;
    }
#if !N_LARGER_THAN_1
#undef i  
#endif

#if 0
  fprintf(stderr, "conv(");
  PRINT_POINT(p, stderr);
  fprintf(stderr, ") = %d\n", sum);
#endif
  return sum;
}
#endif /* BOUNDS_CHECKING */

INLINE_MODIFIER PTR_TO_T ti_ptr(const char *where,
				       const ti_ARRAY x, ti_POINT p) {
  PTR_TO_T result;
  INDEX(result, x.A, ti_conv(where, &x, p));
  return result;
}

INLINE_MODIFIER T ti_read(const char *where,
			  const ti_ARRAY x, ti_POINT p) {
  PTR_TO_T ptr = ti_ptr(where, x, p);
  T result;
  deref(result, ptr);
  return result;
}

#if assignable
INLINE_MODIFIER void ti_write(const char *where,
			      const ti_ARRAY x, ti_POINT p, T value) {
  PTR_TO_T ptr = ti_ptr(where, x, p);
  assign(ptr, value);
}
#endif

INLINE_MODIFIER ti_RECTDOMAIN DOMAIN_METHOD(const ti_ARRAY x)
{
  assertExists(x, "(in domain method)");
  return x.domain;
}

INLINE_MODIFIER ti_ARRAY RESTRICT_METHOD(const ti_ARRAY x, ti_RECTDOMAIN R)
{
  ti_ARRAY n = x;
  assertExists(x, "(in restrict method)");
  n.domain = RECTDOMAIN_INTERSECTION(x.domain, R);
  return n;
}

INLINE_MODIFIER jint CREATOR_METHOD(const ti_ARRAY x)
{
  assertExists(x, "(in creator method)");
  /* creator returns "the lowest numbered process whose demesne contains
     the object pointed to" */
#if EXPLICITLY_STORE_CREATOR
  return COMM_GetProxyProcNumberForBoxNumber(COMM_GetBoxNumberForProcNumber(x.creator));
#else
  return COMM_GetProxyProcNumberForBoxNumber(BOX_PART(x.A));
#endif
}

INLINE_MODIFIER jboolean ISLOCAL_METHOD(const ti_ARRAY x)
{
  assertExists(x, "(in isLocal method)");
#ifndef MEMORY_DISTRIBUTED
  return 1;
#else
  return (COMM_GetBoxNumberForProcNumber(CREATOR_METHOD(x)) == MYBOX);
#endif
}

INLINE_MODIFIER Region REGIONOF_METHOD(const ti_ARRAY x)
{
  RegionId rid;
  assertExists(x, "(in regionOf method)");

  #if GLOBAL_ARRAY
    rid = PG2RegionId(x.A);
  #else
    rid = PL2RegionId(x.A);
  #endif

  if (rid == UNK_REGIONID)
    return NULL;
  else
    return RegionId2Region(rid);
}

#undef POINT_PRINT
#undef N_LARGER_THAN_1
#undef LOCAL_PART
#undef BOX_PART
#undef PROC_PART
#undef INDEX
#undef deref
#undef assign
#undef weak_assign
#undef assignable
